from lura import logs
from lura.run import run

log = logs.get_logger(__name__)

def convert_opts(opts):
  def convert_opt(name, value):
    name = name.replace('_', '-')
    if value in (None, ):
      raise ValueError(f'Invalid value for argument {name}: {value}')
    elif isinstance(value, bool):
      if value:
        value = None
      else:
        return None
    return f'--{name}' if value is None else f'--{name}={value}'
  return [
    opt for opt in (convert_opt(name, val) for name, val in opts.items())
    if opt is not None
  ]

def systemctl(cmd, args, opts, enforce=True, log_level=None):
  log.noise(f'systemctl({cmd}, {args}, {opts})')
  argv = [systemctl.bin, cmd]
  argv.extend(args)
  argv.extend(convert_opts(opts))
  log_level = log_level or systemctl.log_level
  with run.log(log, log_level):
    return run(argv, enforce=enforce)

systemctl.bin = 'systemctl'
systemctl.log_level = log.DEBUG

def journalctl(args, opts, log_level=None):
  log.noise(f'journalctl({args}, {opts})')
  argv = [journalctl.bin]
  argv.extend(args)
  argv.extend(convert_opts(opts))
  log_level = log_level or journalctl.log_level
  with run.log(log, log_level):
    return run(argv)

journalctl.bin = 'journalctl'
journalctl.log_level = log.DEBUG

def start(*args, **opts):
  systemctl('start', args, opts)

def stop(*args, **opts):
  systemctl('stop', args, opts)

def restart(*args, **opts):
  systemctl('restart', args, opts)

def reload(*args, **opts):
  systemctl('reload', args, opts)

def status(*args, **opts):
  return systemctl('status', args, opts, enforce=False).code == 0

def started(svc):
  res = systemctl('status', [svc], {}, enforce=False, log_level=(log.DEBUG))
  return res.code == 0

def daemon_reload(*args, **opts):
  systemctl('daemon-reload', args, opts)

def journal(svc, *args, **opts):
  opts['unit'] = svc
  return str(journalctl(args, opts).stdout)

class Service:

  def __init__(self, name):
    super().__init__()
    self.name = name

  def start(self):
    start(self.name)

  def stop(self):
    stop(self.name)

  def restart(self):
    restart(self.name)

  def reload(self):
    reload(self.name)

  def is_started(self):
    return started(self.name)

  def is_stopped(self):
    return not started(self.name)

  def journal(self, *args, **opts):
    opts['unit'] = self.name
    return str(journalctl(args, opts).stdout)

  started = property(is_started)
  stopped = property(is_stopped)



import os
from lura import fs
from lura import logs
from lura.plates import jinja2

log = logs.get_logger(__name__)

class Installer:

  def __init__(self, name, assets=None):
    super().__init__()
    self.name = name
    self.assets = assets

  def get_directories(self):
    return []

  def get_files(self):
    return []

  def get_assets(self):
    return []

  def get_template_files(self):
    return []

  def get_template_assets(self):
    return []

  def get_symlinks(self):
    return []

  def get_template_env(self):
    return {}

  def get_install_files(self):
    files = []
    files += [file for _, file in self.get_files()]
    files += [file for _, file in self.get_assets()]
    files += [tmpl for _, tmpl in self.get_template_files()]
    files += [tmpl for _, tmpl in self.get_template_assets()]
    files += [symlink for _, symlink in self.get_symlinks()]
    return files

  def create_directory(self, dir):
    if os.path.isdir(dir):
      log.info(f'    {dir} (exists)')
    else:
      log.info(f'    {dir}')
      os.makedirs(dir)

  def create_directories(self):
    dirs = self.get_directories()
    log.info(f'Creating {len(dirs)} directories')
    for dir in dirs:
      self.create_directory(dir)

  def copy_file(self, src, dst):
    if os.path.exists(dst) or os.path.islink(dst):
      log.info(f'    {dst} (overwrite)')
    else:
      log.info(f'    {dst}')
    fs.copy(src, dst)

  def copy_files(self):
    files = self.get_files()
    log.info(f'Copying {len(files)} files')
    for src, dst in files:
      self.copy_file(src, dst)

  def copy_asset(self, src, dst):
    if os.path.exists(dst) or os.path.islink(dst):
      log.info(f'    {dst} (overwrite)')
    else:
      log.info(f'    {dst}')
    self.assets.copy(src, dst)

  def copy_assets(self):
    assets = self.get_assets()
    log.info(f'Copying {len(assets)} assets')
    for src, dst in assets:
      self.copy_asset(src, dst)

  def expand_template_file(self, env, src, dst):
    if os.path.exists(dst) or os.path.islink(dst):
      log.info(f'    {dst} (overwrite)')
    else:
      log.info(f'    {dst}')
    jinja2.expandff(env, src, dst)

  def expand_template_files(self):
    template_files = self.get_template_files()
    log.info(f'Expanding {len(template_files)} template files')
    template_env = self.get_template_env()
    for src, dst in template_files:
      self.expand_template_file(template_env, src, dst)

  def expand_template_asset(self, env, src, dst):
    if os.path.exists(dst) or os.path.islink(dst):
      log.info(f'    {dst} (overwrite)')
    else:
      log.info(f'    {dst}')
    tmpl = self.assets.loads(src)
    jinja2.expandsf(env, tmpl, dst)

  def expand_template_assets(self):
    template_assets = self.get_template_assets()
    log.info(f'Expanding {len(template_assets)} template assets')
    template_env = self.get_template_env()
    for src, dst in template_assets:
      self.expand_template_asset(template_env, src, dst)

  def create_symlink(self, src, dst):
    if os.path.islink(dst) or os.path.exists(dst):
      log.info(f'    {dst} (overwrite)')
      os.unlink(dst)
    else:
      log.info(f'    {dst}')
    os.symlink(src, dst)

  def create_symlinks(self):
    symlinks = self.get_symlinks()
    log.info(f'Creating {len(symlinks)} symlinks')
    for src, dst in symlinks:
      self.create_symlink(src, dst)

  def remove_file(self, path):
    if os.path.exists(path) or os.path.islink(path):
      log.info(f'    {path}')
      os.unlink(path)
    else:
      log.info(f'    {path} (missing)')

  def remove_files(self):
    files = list(reversed(self.install_files))
    log.info(f'Removing {len(files)} installed files')
    for file in files:
      self.remove_file(file)

  def remove_directory(self, path):
    if os.path.isdir(path):
      if len(os.listdir(path)) > 0:
        log.warn(f'    {path} (not empty, keeping)')
      else:
        log.info(f'    {path}')
        os.rmdir(path)
    else:
      log.info(f'    {path} (missing)')

  def remove_directories(self):
    dirs = self.get_directories()
    log.info(f'Removing {len(dirs)} installed directories')
    for path in dirs:
      self.remove_directory(path)

  def on_install_start(self, force):
    pass

  def on_install_finish(self, force):
    pass

  def install(self, force=False):
    log.info(f'Install starting for {self.name}')
    self.on_install_start(force)
    self.create_directories()
    self.copy_files()
    self.copy_assets()
    self.expand_template_files()
    self.expand_template_assets()
    self.create_symlinks()
    self.on_install_finish(force)
    log.info(f'Install finished for {self.name}')

  def on_uninstall_start(self, force):
    pass

  def on_uninstall_finish(self, force):
    pass

  def uninstall(self, force=False):
    log.info(f'Uninstall starting for {self.name}')
    self.on_uninstall_start(force)
    self.remove_files()
    self.remove_directories()
    self.on_uninstall_finish(force)
    log.info(f'Uninstall finished for {self.name}')

  def is_installed(self):
    installed = any(os.path.exists(path) for path in self.install_files)
    if installed:
      log.debug(f'{self.name} is installed')
    else:
      log.debug(f'{self.name} is not installed')
    return installed

  install_files = property(get_install_files)
  installed = property(is_installed)



import jinja2 as _jinja2
import os
from lura.plates import base

class Expander(base.Expander):
  '''
  Expand templates using Jinja2.

  The following extensions are enabled:

  - jinja2.ext.do
  - jinja2.ext.loopcontrols

  The following environment settings are used:

  - trim_blocks
  - lstrip_blocks
  - FileSystemLoader for cwd
  '''

  extensions = [
    'jinja2.ext.do',
    'jinja2.ext.loopcontrols',
  ]

  def __init__(self, cwd=None):
    super().__init__()
    cwd = os.getcwd() if cwd is None else cwd
    self.engine = _jinja2.Environment(
      trim_blocks = True,
      lstrip_blocks = True,
      loader = _jinja2.FileSystemLoader(cwd),
      extensions = self.extensions,
    )

  def expandss(self, env, tmpl):
    return self.engine.from_string(tmpl).render(env)

  def expandsf(self, env, tmpl, dst, encoding=None):
    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(self.engine.from_string(tmpl).render(env))



'SSL RPC services using RPyC.'

import rpyc
from lura import logs
from rpyc.core.service import SlaveService
from rpyc.utils.authenticators import SSLAuthenticator
from rpyc.utils.server import ThreadedServer

log = logs.get_logger(__name__)

Service = SlaveService

def listen(
  service, host, port, key_path, cert_path, sync_timeout, backlog, log=log
):
  name = service.get_service_name()
  log.info(f'Service listening: {name}@{host}:{port}')
  protocol_config = dict(
    allow_all_attrs = True,
    allow_delattr = True,
    allow_public_attrs = True,
    allow_setattr = True,
    logger = log,
    sync_request_timeout = sync_timeout)
  authenticator = SSLAuthenticator(key_path, cert_path)
  server = ThreadedServer(service,
    hostname=host, port=port, protocol_config=protocol_config,
    authenticator=authenticator, backlog=backlog, logger=log)
  server.start()

def patch_close(conn, name):
  conn_close = conn.close
  def close(*args, **kwargs):
    log.info(f'Disconnecting from service: {name}@{conn.host}:{conn.port}')
    conn_close(*args, **kwargs)
    conn.close = conn_close
  conn.close = close

def connect(host, port, key_path, cert_path, sync_timeout, log=log):
  log.info(f'Connecting to service: {host}:{port}')
  protocol_config = dict(
    allow_all_attrs = True,
    allow_delattr = True,
    allow_public_attrs = True,
    allow_setattr = True,
    logger = log,
    sync_request_timeout = sync_timeout)
  conn = rpyc.ssl_connect(
    host, port=port, keyfile=key_path, certfile=cert_path,
    config=protocol_config)
  name = conn.root.get_service_name()
  patch_close(conn, name)
  conn.host = host
  conn.port = port
  conn.service = conn.root
  log.info(f'Connected to service: {name}@{host}:{port}')
  return conn



from string import Template
from lura.plates import base

class Expander(base.Expander):
  'Expand templates using string.Template().'

  def __init__(self):
    super().__init__()

  def expandss(self, env, tmpl):
    return Template(tmpl).substitute(env)

  def expandsf(self, env, tmpl, dst, encoding=None):
    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(Template(tmpl).substitute(env))



import sys
import threading
from lura.attrs import attr
from multiprocessing.pool import ThreadPool

def map(thread_count, func, items, chunksize=None):
  with ThreadPool(thread_count) as p:
    return p.map(func, items, chunksize=chunksize)

def imap(thread_count, func, items, chunksize=1):
  with ThreadPool(thread_count) as p:
    return p.imap(func, items, chunksize=chunksize)

pool = attr(
  map = map,
  imap = imap,
)

class Thread(threading.Thread):

  @classmethod
  def spawn(cls, *args, **kwargs):
    thread = cls(*args, **kwargs)
    thread.start()
    return thread

  def __init__(
    self, group=None, target=None, name=None, args=(), kwargs={}, *,
    daemon=None, reraise=True
  ):
    super().__init__(group=group, name=name, daemon=daemon, target=self.__work)
    self._thread_target = target
    self._thread_args = args
    self._thread_kwargs = kwargs
    self._thread_reraise = reraise
    self._thread_result = None
    self._thread_error = None

  def __work(self):
    try:
      self._thread_result = self._thread_target(
        *self._thread_args, **self._thread_kwargs)
    except Exception:
      self._thread_error = sys.exc_info()
      if self._thread_reraise:
        raise

  @property
  def result(self):
    return self._thread_result

  @property
  def error(self):
    return self._thread_error



import os
from lura import logs
from lura import fs
from lura import run

log = logs.get_logger(__name__)

def convert_opts(opts):
  def convert_opt(name, value):
    name = name.replace('_', '-')
    if value in (None, ):
      raise ValueError(f'Invalid value for git argument {name}: {value}')
    elif isinstance(value, bool):
      if value:
        value = None
      else:
        return None
    return f'--{name}' if value is None else f'--{name}={value}'
  return [
    opt for opt in (convert_opt(name, val) for name, val in opts.items())
    if opt is not None
  ]

def git(cmd, args=[], opts={}, env=None, cwd=None):
  argv = [git.bin, cmd]
  argv.extend(args)
  argv.extend(convert_opts(opts))
  with run.log(log, git.log_level):
    return run(argv, env=env, cwd=cwd)

git.bin = 'git'
git.log_level = log.DEBUG

def clone(*args, env=None, cwd=None, **opts):
  git('clone', args, opts, env, cwd)

def pull(*args, env=None, cwd=None, **opts):
  git('pull', args, opts, env, cwd)

def checkout(*args, env=None, cwd=None, **opts):
  git('checkout', args, opts, env, cwd)

def push(*args, env=None, cwd=None, **opts):
  git('push', args, opts, env, cwd)

class SshKeyEnv(fs.TempDir):

  def __init__(self, key_data=None):
    super().__init__('lura-git', False)
    self.key_data = key_data

  def __enter__(self):
    log.debug('Preparing git ssh authentication')
    temp = super().__enter__()
    key_file = os.path.join(temp, 'id_rsa')
    io.dump(key_file, self.key_data)
    os.chmod(key_file, 0o600)
    ssh_cmd = f"ssh -i '{key_file}' -o StrictHostKeyChecking=no"
    return {'GIT_SSH_COMMAND': ssh_cmd}

class Repo:

  def __init__(self, path, key_data=None):
    super().__init__()
    self.path = path
    self.key_data = key_data

  def clone(self, remote, **opts):
    if self.key_data:
      with SshKeyEnv(self.key_data) as (env):
        clone(remote, self.path, env=env, **opts)
    else:
      clone(remote, (self.path), **opts)

  def pull(self, *args, **opts):
    if self.key_data:
      with SshKeyEnv(self.key_data) as (env):
        pull(args, env=env, cwd=self.path, **opts)
    else:
      pull(args, cwd=self.path, **opts)

  def checkout(self, *args, **opts):
    checkout(args, cwd=self.path, **opts)

  def push(self, *args, **opts):
    if self.key_data:
      with SshKeyEnv(self.key_data) as (env):
        push(args, env=env, cwd=cwd, **opts)
    else:
      push(args, cwd=cwd, **opts)



from abc import abstractmethod

class Expander:

  def __init__(self, *args, **kwargs):
    super().__init__()

  def bind(self, env):
    return BoundExpander(self, env)

  @abstractmethod
  def expandss(self, env, tmpl):
    'Expand a `str` to a `str`.'

    pass

  @abstractmethod
  def expandsf(self, env, tmpl, dst, encoding=None):
    'Expand a `str` to a file.'

    pass

  def expandfs(self, env, src, encoding=None):
    'Expand a file to a `str`.'

    with open(src, 'r', encoding=encoding) as fd:
      tmpl = fd.read()
    return self.expandss(env, tmpl)

  def expandff(self, env, src, dst, encoding=None):
    'Expand a file to a file.'

    with open(src, 'r', encoding=encoding) as fd:
      tmpl = fd.read()
    self.expandsf(env, tmpl, dst, encoding=encoding)

class BoundExpander:
  'Binds an expander to an environment.'

  def __init__(self, expander, env):
    super().__init__()
    self.expander = expander
    self.env = env

  def expandss(self, tmpl):
    return self.expander.expandss(self.env, tmpl)

  def expandsf(self, tmpl, dst):
    return self.expander.expandsf(self.env, tmpl, dst)

  def expandfs(self, src):
    return self.expander.expandfs(self.env, src)

  def expandff(self, src, dst):
    return self.expander.expandff(self.env, src, dst)



from lura import logs
from lura.run import run

log = logs.get_logger(__name__)

def convert_opts(opts):
  def convert_opt(name, value):
    name = name.replace('_', '-')
    if value in (None, ):
      raise ValueError(f'Invalid value for docker argument {name}: {value}')
    elif isinstance(value, bool):
      if value:
        value = None
      else:
        return None
    else:
      value = str(value)
    if value is None:
      return (f'--{name}',)
    else:
      return (f'--{name}', value)
  argv = []
  for name, value in opts.items():
    converted = convert_opt(name, value)
    if converted is not None:
      argv.extend(converted)
  return argv

def docker(cmd, args=[], opts={}, cwd=None, enforce=True):
  argv = [docker.bin, cmd]
  argv.extend(args)
  argv.extend(convert_opts(opts))
  with run.log(log, docker.log_level):
    return run(argv, cwd=cwd, enforce=enforce)

docker.bin = 'docker'
docker.log_level = log.DEBUG

def build(*args, cwd=None, **opts):
  docker('build', args, opts, cwd)

def tag(*args, **opts):
  docker('tag', args, opts)

def push(*args, **opts):
  docker('push', args, opts)

def images(*args, **opts):
  opts['format'] = '{{json .}}'
  return docker('images', args, opts).stdout.json()

def rmi(*args, enforce=True, **opts):
  docker('rmi', args, opts, enforce=enforce)



import os
import sys
from abc import abstractmethod
from lura.hash import hashs

class Format:

  def __init__(self, *args, **kwargs):
    super().__init__()

  @abstractmethod
  def loads(self, data):
    pass

  @abstractmethod
  def loadf(self, src, encoding=None):
    pass

  @abstractmethod
  def loadfd(self, fd):
    pass

  @abstractmethod
  def dumps(self, data):
    pass

  @abstractmethod
  def dumpf(self, data, dst, encoding=None):
    pass

  @abstractmethod
  def dumpfd(self, data, fd):
    pass

  def print(self, data, *args, **kwargs):
    print(self.dumps(data).rstrip(), *args, **kwargs)



from lura import logs
from lura.run import run
log = logs.get_logger(__name__)

def convert_opts(opts):
  def convert_opt(name, value):
    name = name.replace('_', '-')
    if value in (None, ):
      raise ValueError(
        f'Invalid value for docker-compose argument {name}: {value}')
    elif isinstance(value, bool):
      if value:
        value = None
      else:
        return None
    else:
      value = str(value)
    return f'--{name}' if value is None else f'--{name}={value}'
  argv = []
  for name, value in opts.items():
    converted = convert_opt(name, value)
    if converted is not None:
      argv.extend(converted)
  return argv

def docker_compose(cmd, file=None, args=[], opts={}, cwd=None):
  argv = [docker_compose.bin]
  if file:
    argv.extend(('-f', file))
  argv.append(cmd)
  argv.extend(args)
  argv.extend(convert_opts(opts))
  with run.log(log, docker_compose.log_level):
    return run(argv, cwd=cwd)

docker_compose.bin = 'docker-compose'
docker_compose.log_level = log.DEBUG

def pull(*args, file=None, cwd=None, **opts):
  docker_compose('pull', file, args, opts, cwd)



import io
import os
import shlex
import subprocess as subp
import sys
import threading
from contextlib import contextmanager
from deepmerge import always_merger
from lura import LuraError
from lura import formats
from lura import logs
from lura import threads
from lura.attrs import attr
from lura.attrs import deepcopy
from lura.attrs import ottr
from lura.sudo import SudoHelper
from lura.sudo import shell_path

log = logs.get_logger(__name__)

merge = always_merger.merge
shjoin = subp.list2cmdline

def tee(source, targets, cond=lambda: True):
  '`readline` from one source and `write` to many targets.'

  while cond():
    line = source.readline()
    if line == '':
      break
    for target in targets:
      target.write(line)

class Tee(threads.Thread):
  '`readline` from one source and `write` to many targets in a thread.'

  def __init__(self, source, targets):
    super().__init__(target=self.work, name='Tee')
    self.source = source
    self.targets = targets
    self._work = False

  def work(self):
    self._work = True
    tee(self.source, self.targets, lambda: self._work is True)

  def stop(self):
    self._work = False

class LogWriter:
  'A file-like object which logs data written to it.'

  # this is used to implement the `run.log(...)` context manager

  def __init__(self, logger, level, tag):
    super().__init__()
    self.log = logger.method_for_level(level)
    self.tag = tag

  def write(self, line):
    self.log(f'[{self.tag}] {line.rstrip()}')

class Result:
  'The value returned by `run()`.'

  members = ('args', 'argv', 'code', 'stdout', 'stderr')

  def __init__(self, context, code, stdout, stderr):
    super().__init__()
    self.args = context.args
    self.argv = context.argv
    self.code = code
    self.stdout = stdout
    self.stderr = stderr
    self.context = context
    del context['args']
    del context['argv']
    if context.sudo_password:
      context.sudo_password = '<scrubbed>'

  def as_dict(self):
    return ottr((_, getattr(self, _)) for _ in self.members)

  def format(self, fmt='yaml'):
    tag = 'run.{}'.format(type(self).__name__.lower())
    return formats.for_ext(fmt).dumps({tag: self.as_dict()})

  def print(self, fmt='yaml', file=None):
    file = sys.stdout if file is None else file
    file.write(self.format(fmt=fmt))

class Error(LuraError):
  'The error raised by `run()` when a process exits with the wrong code.'

  def __init__(self, result):
    msg = 'Process expected exit code {} but received code {}{}{}'.format(
      result.context.enforce_code, result.code, os.linesep, result.format())
    super().__init__(msg)
    self.result = result

class Run(threading.local):
  'The implementation of `run()`.'

  # kwargs that we accept
  _kwargs = (
    'pty', 'env', 'env_replace', 'cwd', 'shell', 'stdin', 'stdout', 'stderr',
    'encoding', 'enforce', 'enforce_code', 'sudo', 'sudo_user', 'sudo_group',
    'sudo_password', 'sudo_login'
  )

  # default kwarg values
  defaults = dict(
    pty           = False,
    env           = None,
    env_replace   = False,
    cwd           = None,
    shell         = False,
    stdin         = None,
    stdout        = [],
    stderr        = [],
    encoding      = sys.getdefaultencoding(),
    enforce       = True,
    enforce_code  = 0,
    sudo          = False,
    sudo_user     = None,
    sudo_group    = None,
    sudo_password = None,
    sudo_login    = False,
  )

  shell = shell_path()

  # for stdio threads
  join_timeout = 0.5

  def __init__(self):
    super().__init__()
    # context managers set their values here
    self.context = {}

  def _subprocess(self, ctx):
    'Run with `subprocess.Popen`.'

    outbuf, errbuf = io.StringIO(), io.StringIO()
    stdout = [outbuf] + ctx.stdout
    stderr = [errbuf] + ctx.stderr
    argv = ctx.args if ctx.shell else ctx.argv
    proc = None
    threads = ()
    try:
      proc = subp.Popen(
        argv, env=ctx.env, cwd=ctx.cwd, shell=ctx.shell, stdout=subp.PIPE,
        stderr=subp.PIPE, stdin=ctx.stdin, encoding=ctx.encoding)
      if ctx.sudo:
        ctx.sudo_helper.wait()
      threads = (Tee.spawn(proc.stdout, stdout), Tee.spawn(proc.stderr, stderr))
      code = proc.wait()
      for thread in threads:
        thread.join()
      threads = ()
      return self.Result(ctx, code, outbuf.getvalue(), errbuf.getvalue())
    finally:
      for thread in threads:
        thread.stop()
        thread.join(self.join_timeout)
        if thread.is_alive():
          log.debug(f'Failed joining tee thread: {thread}')
      outbuf.close()
      errbuf.close()
      if proc:
        proc.kill()

  def _ptyprocess(self, ctx):
    'Run with `PtyProcessUnicode`.'

    if ctx.stdin:
      raise NotImplementedError('stdin not implemented for pty processes')
    if ctx.shell:
      ctx.argv = [self.shell, '-c', ctx.args]
      ctx.args = shjoin(ctx.argv)
    outbuf = io.StringIO()
    stdout = [outbuf] + ctx.stdout
    proc = None
    try:
      proc = PtyProcessUnicode.spawn(argv, env=ctx.env, cwd=ctx.cwd)
      if ctx.sudo:
        ctx.sudo_helper.wait()
      reader = attr(readline=lambda: f'{proc.readline()[:-2]}{os.linesep}')
      try:
        tee(reader, stdout)
      except EOFError:
        pass
      return self.Result(ctx, proc.wait(), outbuf.getvalue(), '')
    finally:
      outbuf.close()
      if proc:
        try:
          proc.kill(9)
        except Exception:
          log.debug('Failed killing pty process', exc_info=True)

  def _check_args(self, kwargs):
    'Validate arg names in `self.defaults`, `self.context`, and `kwargs`.'

    def check_args(name, bundle):
      for arg in bundle:
        if arg not in self._kwargs:
          raise ValueError(f'Unknown argument in {name}: {arg}={bundle[arg]}')
    check_args('defaults', self.defaults)
    check_args('context', self.context)
    check_args('kwargs', kwargs)

  def _build_context(self, kwargs):
    defaults = deepcopy(self.defaults)
    context = deepcopy(self.context)
    return merge(defaults, merge(context, kwargs))

  def _env_setup(self, ctx):
    'Setup the environment the process will run in.'

    if ctx.env is None:
      ctx.env = {}
    if not ctx.env_replace:
      env = {}
      env.update(os.environ)
      env.update(ctx.env)
      ctx.env = env

  def _sudo_setup(self, ctx, argv):
    'Setup sudo, returning the new sudo argv.'

    ctx.sudo_helper = SudoHelper(
      user=ctx.sudo_user, group=ctx.sudo_group, password=ctx.sudo_password,
      login=ctx.sudo_login)
    argv, askpass_path = ctx.sudo_helper.prepare(argv)
    if askpass_path:
      ctx.env['SUDO_ASKPASS'] = askpass_path
    return argv

  def _sudo_cleanup(self, ctx):
    ctx.sudo_helper.cleanup()
    del ctx['sudo_helper']

  def _args_argv(self, ctx, argv):
    'ctx.args is always a string, ctx.argv is always a list.'

    if isinstance(argv, str):
      ctx.args = argv
      ctx.argv = shlex.split(argv)
    else:
      ctx.args = shjoin(argv)
      ctx.argv = argv

  def run(self, argv, **kwargs):
    'Entry point for `run()`.'

    self._check_args(kwargs)
    ctx = self._build_context(kwargs)
    self._env_setup(ctx)
    if ctx.sudo:
      argv = self._sudo_setup(ctx, argv)
    try:
      self._args_argv(ctx, argv)
      if ctx.pty:
        result = self._ptyprocess(ctx)
      else:
        result = self._subprocess(ctx)
      if ctx.enforce and result.code != ctx.enforce_code:
        raise self.Error(result)
      return result
    finally:
      if ctx.sudo:
        self._sudo_cleanup(ctx)

  __call__ = run

  @contextmanager
  def quash(self):
    'Do not enforce exit code while in this context.'

    old = self.context
    new = {'enforce': False}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def enforce(self, enforce_code=0):
    'Enforce the given exit code while in this context.'

    old = self.context
    new = {'enforce': True, 'enforce_code': enforce_code}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def cwd(self, cwd):
    'Run in directory `cwd` while in this context.'

    old = self.context
    new = {'cwd': cwd}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def shell(self):
    "Run all commands with the user's shell."

    old = self.context
    new = {'shell': True}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def sudo(
    self, sudo_user=None, sudo_group=None, sudo_password=None, sudo_login=None
  ):
    'Run with sudo while in this context.'

    old = self.context
    new = dict(
      sudo = True,
      sudo_user = sudo_user,
      sudo_group = sudo_group,
      sudo_password = sudo_password,
      sudo_login = sudo_login,
    )
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def log(self, logger, log_level=log.DEBUG):
    'Send stdout and stderr to a logger while in this context.'

    old = self.context
    new = {
      'stdout': [LogWriter(logger, log_level, 'out')],
      'stderr': [LogWriter(logger, log_level, 'err')],
    }
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

Run.Error = Error
Run.Result = Result

run = Run()



'System abstractions for ochestration in ssh and local flavors.'

import os
from abc import abstractmethod
from contextlib import contextmanager
from lura import fs
from lura import run
from lura import ssh
from shlex import quote

class System:

  def __init__(self):
    super().__init__()
    self.use_sudo = False

  @abstractmethod
  def put(self, src, dst):
    pass

  @abstractmethod
  def get(self, dst, src):
    pass

  @abstractmethod
  def run(
    self, argv, shell=False, pty=False, env=None, replace_env=False,
    encoding=None, stdin=None, stdout=None, stderr=None, enforce=True
  ):
    pass

  @contextmanager
  def sudo(self):
    orig = self.use_sudo
    self.use_sudo = True
    try:
      yield self
    finally:
      self.use_sudo = orig

  @contextmanager
  def nosudo(self):
    orig = self.use_sudo
    self.use_sudo = False
    try:
      yield self
    finally:
      self.use_sudo = orig

  @contextmanager
  def tempdir(self, prefix=None):
    prefix = prefix or 'lura.system.'
    path = self.run(f'mktemp -d {quote(prefix)}.' + 'X' * 12).stdout.rstrip()
    try:
      yield path
    finally:
      self.rmrf(path)

  def read(self, path):
    with fs.TempDir(prefix='lura-system-read.') as temp_dir:
      dst = f'{temp_dir}/{os.path.basename(path)}'
      self.get(path, dst)
      return fs.loads(dst)

  def write(self, path, data):
    with fs.TempDir(prefix='lura-system-write.') as temp_dir:
      src = f'{temp_dir}/{os.path.basename(path)}'
      fs.dumps(src, data)
      self.put(src, path)

  def whoami(self):
    return self.run('whoami').stdout.rstrip()

  def ls(self, path, long=False):
    argv = f'/bin/ls -a --indicator-style=none {quote(path)}|cat'
    files = self.run(argv).stdout.rstrip().split('\n')
    files = [_ for _ in files if _ not in ('.', '..')]
    if long:
      return [os.path.join(path, _) for _ in files]
    else:
      return files

  def cpf(self, src, dst, preserve=False):
    cp = ['cp', '-f']
    if preserve:
      cp.append('--preserve=all')
    cp.extend((quote(src), quote(dst)))
    cp = ' '.join(cp)
    self.run(cp)

  def cprf(self, src, dst, preserve=False):
    argv = ['cp', '-rf']
    if preserve:
      argv.append('--preserve=all')
    argv.extend((quote(src), quote(dst)))
    argv = ' '.join(argv)
    self.run(argv)

  def mv(self, src, dst):
    self.run(f'mv -f {quote(src)} {quote(dst)}')

  def rmf(self, path):
    self.run(f'rm -f {quote(path)}')

  def rmrf(self, path):
    self.run(f'rm -rf {quote(path)}')

  def access(self, path, flag):
    raise NotImplementedError(':(')

  def exists(self, path):
    res = self.run(f'test -e {quote(path)}', enforce=False)
    return not bool(res.return_code)

  def isfile(self, path):
    res = self.run(f'test -f {quote(path)}', enforce=False)
    return not bool(res.return_code)

  def isdir(self, path):
    res = self.run(f'test -d {quote(path)}', enforce=False)
    return not bool(res.return_code)

  def islink(self, path):
    res = self.run(f'test -L {quote(path)}', enforce=False)
    return not bool(res.return_code)

  def isfifo(self, path):
    raise NotImplementedError()

  def which(self, *names, error=False):
    names = ' '.join(quote(_) for _ in names)
    return self.run(f'which {names}', enforce=False).stdout.rstrip()

  def chmod(self, path, mode, recurse=False):
    argv = ['chmod']
    if recurse:
      argv.append('-R')
    argv.extend((mode, quote(path)))
    argv = ' '.join(argv)
    self.run(argv)

  def chown(self, path, spec, recurse=False):
    argv = ['chown']
    if recurse:
      argv.append('-R')
    argv.extend((spec, quote(path)))
    argv = ' '.join(argv)
    self.run(argv)

  def chgrp(self, path, group, recurse=False):
    argv = ['chgrp']
    if recurse:
      argv.append('-R')
    argv.extend((group, quote(path)))
    argv = ' '.join(argv)
    self.run(argv)

  def touch(self, path):
    self.run(f'touch {quote(path)}')

  def mkdir(self, dir):
    if self.isdir(dir):
      return
    self.run(f'mkdir {quote(dir)}')

  def mkdirp(self, dir):
    if self.isdir(dir):
      return
    self.run(f'mkdir -p {quote(dir)}')

  def hostname(self):
    raise NotImplementedError()

  @property
  def shell(self):
    return self.run('echo $0').stdout.rstrip()

class Local(System):

  def __init__(self, sudo_password=None, sudo_user=None):
    super().__init__()
    self.sudo_password = sudo_password
    self.sudo_user = sudo_user

  def put(self, src, dst):
    self.cpf(src, dst)

  def get(self, dst, src):
    self.cpf(src, dst)

  def run(self, *args, **kwargs):
    if self.use_sudo:
      kwargs['sudo'] = True
      if self.sudo_password:
        kwargs['sudo_password'] = self.sudo_password
      if self.sudo_user:
        kwargs['sudo_user'] = self.sudo_user
    kwargs['shell'] = True
    res = run(*args, **kwargs)
    res.return_code = res.code # like fabric
    return res

  def read(self, path):
    return fs.loads(path)

  def write(self, path, data):
    return fs.writes(path, data)

class Ssh(System):

  def __init__(self, client, sudo_user=None):
    super().__init__()
    self.client = client
    self.sudo_user = sudo_user

  # we want put and get to honor self.use_sudo. fabric's put and get are not
  # sudoable, so far as I can tell. the crude methods below should work in the
  # sudo and non-sudo cases, but this should be improved.

  def put(self, src, dst):
    with self.tempdir() as temp_dir:
      tmpdst = f'{temp_dir}/{os.path.basename(dst)}'
      self.client.put(src, tmpdst)
      self.cpf(tmpdst, dst)

  def get(self, dst, src):
    with self.nosudo():
      whoami = self.whoami()
    with self.tempdir() as temp_dir:
      tmpsrc = f'{temp_dir}/{os.path.basename(src)}'
      self.cpf(src, tmpsrc)
      self.chown(temp_dir, whoami, recurse=True)
      self.client.get(dst, tmpsrc)

  def run(self, *args, **kwargs):
    if self.use_sudo:
      return self.client.sudo(
        *args, user=self.sudo_user, **kwargs)
    else:
      return self.client.run(*args, **kwargs)



'SSH client with sudo support.'

import fabric
import sys
from invoke import Responder
from lura import logs
from subprocess import list2cmdline as shjoin

log = logs.get_logger(__name__)

class Client:

  log_level = log.DEBUG

  def __init__(
    self, host, port=22, user=None, password=None, key_file=None,
    passphrase=None, timeout=60.0, auth_timeout=60.0, sudo_password=None
  ):
    # FIXME accept key data from buffer
    super().__init__()
    self.host = host
    self.port = port
    self.user = user
    self.timeout = timeout
    self.conn_kwargs = {
      'key_filename': key_file,
      'passphrase': passphrase,
      'auth_timeout': auth_timeout,
    }
    self.sudo_password = sudo_password
    self.conn = None

  def __enter__(self):
    self.connect()
    return self

  def __exit__(self, *exc_info):
    self.close()

  def _log(self, msg):
    _log = getattr(log, logs.get_level_name(self.log_level).lower())
    _log(f'[{self.host}] {msg}')

  def connect(self):
    self._log('Connecting')
    overrides = {}
    if self.sudo_password:
      overrides['sudo'] = {'password': self.sudo_password}
    config = fabric.Config(overrides=overrides)
    self.conn = fabric.Connection(
      host=self.host, user=self.user, port=self.port,
      connect_timeout=self.timeout, connect_kwargs=self.conn_kwargs,
      config=config)
    self._log('Connected')

  def is_closed(self):
    return self.conn is None

  def close(self):
    try:
      self._log('Closing')
      self.conn.close()
    except Exception:
      log.exception('Unhandled exceptionat close()')
    self.conn = None
    self._log('Closed')
    self.host = None

  disconnect = close

  def put(self, src, dst):
    self._log(f'put {src} -> {dst}')
    self.conn.put(src, remote=dst)

  def get(self, dst, src):
    self._log(f'get {dst} -> {src}')
    self.conn.get(src, local=dst)

  def run(
    self, argv, shell=False, pty=False, env={}, replace_env=False,
    encoding=None, stdin=None, stdout=None, stderr=None, enforce=True
  ):
    if not isinstance(argv, str):
      argv = shjoin(argv)
    self._log(f'run {argv}')
    return self.conn.run(
      argv, shell=shell, pty=pty, env=env, replace_env=replace_env,
      encoding=encoding, in_stream=stdin, out_stream=stdout,
      err_stream=stderr, warn=not enforce, hide=True)

  def sudo(
    self, argv, shell=False, pty=False, env={}, replace_env=False,
    encoding=None, stdin=None, stdout=None, stderr=None, enforce=True,
    user=None
  ):
    if not isinstance(argv, str):
      argv = shjoin(argv)
    self._log(f'sudo {argv}')
    return self.conn.sudo(
      argv, shell=shell, pty=pty, env=env, replace_env=replace_env,
      encoding=encoding, in_stream=stdin, out_stream=stdout,
      err_stream=stderr, warn=not enforce, hide=True,
      user=user)

  def forward(self, lport, rport, lhost=None):
    return self.conn.forward_local(lport, rport, lhost)



import logging
import logging.config
import os
import time
from collections import defaultdict
from io import StringIO
from lura.attrs import attr

class ExtraInfoFilter(logging.Filter):
  '''
  Provides additional fields to log records:

  - `short_name` a reasonably short name
  - `shortest_name` the shortest reasonably useful name
  - `short_levelname` a symbol associated with a level
  - `run_time` number of seconds the logging system has been initialized
  '''

  initialized = time.time()

  map_short_level = defaultdict(
    lambda: '=',
    DEBUG    = '+',
    INFO     = '|',
    WARNING  = '>',
    ERROR    = '*',
    CRITICAL = '!',
  )

  def filter(self, record):
    _ = record.name.split('.')
    record.short_name = '.'.join(_[-2:])
    record.shortest_name = _[-1]
    record.short_levelname = self.map_short_level.get(record.levelname)
    record.run_time = time.time() - self.initialized
    return True

class MultiLineFormatter(logging.Formatter):
  '''
  Format messages containing lineseps as though each line were a log
  message.
  '''

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def format(self, record):
    if not isinstance(record.msg, str) or os.linesep not in record.msg:
      record.message = super().format(record)
      return record.message
    msg = record.msg
    with StringIO() as buf:
      for line in record.msg.split(os.linesep):
        record.msg = line
        buf.write(super().format(record) + os.linesep)
      record.message = buf.getvalue().rpartition(os.linesep)[0]
    record.msg = msg
    return record.message

class Logging:
  '''
  An API for the configuration and maintenance of an application's or
  library's root logger.

  `MultiLineFormatter` and `ExtraInfoFilter` are used by default.
  '''

  # for convenience
  NOTSET   = logging.NOTSET
  DEBUG    = logging.DEBUG
  INFO     = logging.INFO
  WARNING  = logging.WARNING
  WARN     = logging.WARN
  ERROR    = logging.ERROR
  CRITICAL = logging.CRITICAL
  FATAL    = logging.FATAL

  # log format presets
  formats = attr(
    bare    = '%(message)s',
    classic = '%(asctime)s %(levelname)-8s %(short_name)s %(message)s',
    hax     = '%(run_time)-8.3f %(short_name)19s %(short_levelname)s %(message)s',
    runtime = '%(run_time)-12.3f %(message)s',
    verbose = '%(asctime)s %(run_time)12.3f %(name)s %(short_levelname)s %(message)s',
  )

  default_datefmt = '%Y-%m-%d %H:%M:%S'

  def __init__(
    self,
    std_logger,
    std_format = None,
    std_datefmt = None,
  ):
    super().__init__()
    self.std_logger = std_logger
    self.std_format = std_format or self.formats.bare
    self.std_datefmt = std_datefmt or self.default_datefmt
    self.config = None
    self.configure()

  def configure(self):
    if self.config is not None:
      return
    self._config_logging()
    self._config_levels()

  def _config_logging(self):
    self.config = self._build_config()
    logging.config.dictConfig(self.config)
    for level, name in logging._levelToName.items():
      setattr(logging.Logger, name, level)

  def _config_levels(self):
    if 'NOISE' not in logging._levelToName.values():
      self.add_level('NOISE', 5, ':')

  def _build_config(self):
    import yaml
    config = yaml.safe_load(f'''
      version: 1
      disable_existing_loggers: false
      filters:
        short_name: {{}}
      formatters:
        standard:
          format: '{self.std_format}'
          datefmt: '{self.std_datefmt}'
      handlers:
        stderr:
          class: logging.StreamHandler
          stream: ext://sys.stderr
          filters: ['short_name']
          formatter: standard
      loggers:
        {self.std_logger}:
          handlers: ['stderr']
          level: INFO
    ''')
    config['filters']['short_name']['()'] = ExtraInfoFilter
    config['formatters']['standard']['()'] = MultiLineFormatter
    return config

  def _build_logger_log_method(self, level):
    def log_level(self, msg, *args, **kwargs):
      if self.isEnabledFor(level):
        self._log(level, msg, args, **kwargs)
    return log_level

  def get_logger(self, name=None):
    return logging.getLogger(name or self.std_logger)

  getLogger = get_logger

  def get_level(self):
    'Get standard logger log level.'

    return self.get_logger().getEffectiveLevel()

  def set_level(self, level):
    'Set standard logger log level.'

    if isinstance(level, str):
      level = getattr(self, level.upper())
    logger = self.get_logger()
    logger.setLevel(level)

  def get_level_name(self, number):
    return logging._levelToName[number]

  def add_level(self, name, number, short_name=None):
    if number in logging._levelToName.keys():
      msg = 'Log level number {} already used by {}'
      raise ValueError(msg.format(number, self.get_level_name(number)))
    elif name in logging._levelToName.values():
      raise ValueError(f'Log level name {name} already in use')
    logging.addLevelName(number, name)
    setattr(logging, name, number)
    setattr(logging.Logger, name, number)
    setattr(type(self), name, number)
    setattr(
      logging.Logger, name.lower(), self._build_logger_log_method(number))
    if short_name is not None:
      ExtraInfoFilter.map_short_level[name] = short_name

  def set_format(self, format, datefmt=None, logger=None):
    'Sets the format for all handlers of a logger.'

    datefmt = datefmt or self.default_datefmt
    logger = logger or self.std_logger
    logger = self.get_logger(logger)
    formatter = MultiLineFormatter(format, datefmt)
    for handler in logger.handlers:
      handler.setFormatter(formatter)

  def add_file_handler(self, path, level=None, logger=None, append=True):
    logger = logger or self.std_logger
    if isinstance(path, str):
      if not append and os.path.isfile(path):
        os.unlink(path)
      handler = logging.FileHandler(path)
    else:
      handler = logging.StreamHandler(path)
    if level is not None:
      if isinstance(level, str):
        level = getattr(self, level)
      handler.setLevel(level)
    formatter = MultiLineFormatter(*self.formats.verbose)
    handler.setFormatter(formatter)
    handler.addFilter(ExtraInfoFilter())
    self.get_logger(logger).addHandler(handler)
    return handler

  def remove_handler(self, handler, logger=None):
    logger = logger or self.std_logger
    self.get_logger(logger).removeHandler(handler)

def method_for_level(self, level):
  name = logging._levelToName[level].lower()
  return getattr(self, name)

logging.Logger.method_for_level = method_for_level



import sys
import shutil
import tempfile
from lura import logs

log = logs.get_logger(__name__)

def load(path, encoding=None):
  with open(path, 'rb', encoding=encoding) as pathf:
    return pathf.read()

def loads(path, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  with open(path, 'r', encoding=encoding) as pathf:
    return pathf.read()

def dump(path, data, encoding=None):
  with open(path, 'wb', encoding=encoding) as pathf:
    pathf.write(data)

def dumps(path, data, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  with open(path, 'w', encoding=encoding) as pathf:
    pathf.write(data)

class TempDir:

  def __init__(self, suffix=None, prefix=None, dir=None, keep=False):
    super().__init__()
    self._tempdir_suffix = suffix
    self._tempdir_prefix = prefix
    self._tempdir_root = dir
    self._tempdir_keep = keep
    self._tempdir_dir = None

  def __enter__(self):
    self._tempdir_dir = tempfile.mkdtemp(
      suffix=self._tempdir_suffix, prefix=self._tempdir_prefix,
      dir=self._tempdir_root)
    return self._tempdir_dir

  def __exit__(self, *exc_info):
    if self._tempdir_keep:
      log.warn(f'Keeping temporary directory: {self._tempdir_dir}')
    else:
      shutil.rmtree(self._tempdir_dir)
    self._tempdir_dir = None
